from __future__ import annotations

from .base_transport import BaseTransport
from ..enums import ConnectionRole
from ..packets import BasePacket, QuickAckPacket, ErrorPacket, MessagePacket


class AbridgedTransport(BaseTransport):
    def read(self) -> BasePacket | None:
        if self.rx_buffer.size() < 4:
            return

        length = self.rx_buffer.peekexactly(1)[0]
        is_quick_ack = length & 0x80 == 0x80
        length &= 0x7F

        if is_quick_ack and self.our_role == ConnectionRole.CLIENT:
            return QuickAckPacket(self.rx_buffer.readexactly(4)[::-1])

        big_length = length & 0x7F == 0x7F
        if big_length:
            length = int.from_bytes(self.rx_buffer.peekexactly(3, 1), "little")

        length *= 4
        if self.rx_buffer.size() < (length + 4 if big_length else 1):
            return

        self.rx_buffer.readexactly(4 if big_length else 1)
        data = self.rx_buffer.readexactly(length)
        if len(data) == 4:
            return ErrorPacket(int.from_bytes(data, "little", signed=True))

        return MessagePacket.parse(data, is_quick_ack)

    def write(self, packet: BasePacket) -> None:
        data = packet.write()
        if isinstance(packet, QuickAckPacket):
            self.tx_buffer.write(data[::-1])
            return

        length = (len(data) + 3) // 4

        if length >= 0x7F:
            self.tx_buffer.write(b"\x7f")
            self.tx_buffer.write(length.to_bytes(3, byteorder="little"))
        else:
            self.tx_buffer.write(length.to_bytes(1, byteorder="little"))

        self.tx_buffer.write(data)

    def has_packet(self) -> bool:
        if self.rx_buffer.size() < 4:
            return False
        length = self.rx_buffer.peekexactly(1)[0]
        if length & 0x80 == 0x80:
            return True
        length &= 0x7F

        length_size = 1
        if length & 0x7F == 0x7F:
            length_size = 4
            length = int.from_bytes(self.rx_buffer.peekexactly(3, 1), "little")

        length *= 4
        return self.rx_buffer.size() >= (length + length_size)



from functools import partial
from hashlib import sha256, sha1

try:
    import tgcrypto
except ImportError:  # pragma: no cover
    tgcrypto = None
try:
    import pyaes
except ImportError:  # pragma: no cover
    pyaes = None


if tgcrypto is None and pyaes is None:  # pragma: no cover
    raise ImportWarning(
        "Expected at least one or (tgcrypto, pyaes) to be installed. "
        "You wont be able to use obfuscated transports or mtproto messages encryption/decryption."
    )


if tgcrypto is not None:
    _ctr256_encrypt = tgcrypto.ctr256_encrypt
    _ctr256_decrypt = tgcrypto.ctr256_decrypt
    _ige256_encrypt = tgcrypto.ige256_encrypt
    _ige256_decrypt = tgcrypto.ige256_decrypt
elif pyaes is not None:
    # https://github.com/pyrogram/pyrogram/blob/39694a29497aee87d6ee91155e9b7570b7849aa9/pyrogram/crypto/aes.py#L105
    def ctr(data: bytes, key: bytes, iv: bytearray, state: bytearray) -> bytes:
        cipher = pyaes.AES(key)

        out = bytearray(data)
        chunk = cipher.encrypt(iv)

        for i in range(0, len(data), 16):
            for j in range(0, min(len(data) - i, 16)):
                out[i + j] ^= chunk[state[0]]

                state[0] += 1

                if state[0] >= 16:
                    state[0] = 0

                if state[0] == 0:
                    for k in range(15, -1, -1):
                        try:
                            iv[k] += 1
                            break
                        except ValueError:
                            iv[k] = 0

                    chunk = cipher.encrypt(iv)

        return out


    def xor(a: bytes, b: bytes) -> bytes:
        return int.to_bytes(
            int.from_bytes(a, "big") ^ int.from_bytes(b, "big"),
            len(a),
            "big",
        )


    # https://github.com/pyrogram/pyrogram/blob/39694a29497aee87d6ee91155e9b7570b7849aa9/pyrogram/crypto/aes.py#L85
    def ige(data: bytes, key: bytes, iv: bytes, encrypt: bool) -> bytes:
        cipher = pyaes.AES(key)

        iv_1 = iv[:16]
        iv_2 = iv[16:]

        data = [data[i: i + 16] for i in range(0, len(data), 16)]

        if encrypt:
            for i, chunk in enumerate(data):
                iv_1 = data[i] = xor(cipher.encrypt(xor(chunk, iv_1)), iv_2)
                iv_2 = chunk
        else:
            for i, chunk in enumerate(data):
                iv_2 = data[i] = xor(cipher.decrypt(xor(chunk, iv_2)), iv_1)
                iv_1 = chunk

        return b"".join(data)


    _ctr256_encrypt = ctr
    _ctr256_decrypt = ctr
    _ige256_encrypt = partial(ige, encrypt=True)
    _ige256_decrypt = partial(ige, encrypt=False)
else:  # pragma: no cover
    def _no_crypto_library(*args, **kwargs):
        raise RuntimeError(
            "To use obfuscated transports and mtproto messages encryption/decryption, "
            "you need to install either pyaes or tgcrypto."
        )


    _ctr256_encrypt = _no_crypto_library
    _ctr256_decrypt = _no_crypto_library
    _ige256_encrypt = _no_crypto_library
    _ige256_decrypt = _no_crypto_library


def ctr256_encrypt(data: bytes, key: bytes, iv: bytearray, state: bytearray) -> bytes:
    return _ctr256_encrypt(data, key, iv, state)


def ctr256_decrypt(data: bytes, key: bytes, iv: bytearray, state: bytearray) -> bytes:
    return _ctr256_decrypt(data, key, iv, state)


def ige256_encrypt(data: bytes, key: bytes, iv: bytes) -> bytes:
    return _ige256_encrypt(data, key, iv)


def ige256_decrypt(data: bytes, key: bytes, iv: bytes) -> bytes:
    return _ige256_decrypt(data, key, iv)


def kdf(auth_key: bytes, msg_key: bytes, from_client: bool) -> tuple:
    # taken from pyrogram, mtproto.py
    x = 0 if from_client else 8

    sha256_a = sha256(msg_key + auth_key[x:x + 36]).digest()
    sha256_b = sha256(auth_key[x + 40:x + 76] + msg_key).digest()  # 76 = 40 + 36

    aes_key = sha256_a[:8] + sha256_b[8:24] + sha256_a[24:32]
    aes_iv = sha256_b[:8] + sha256_a[8:24] + sha256_b[24:32]

    return aes_key, aes_iv


def kdf_v1(auth_key: bytes, msg_key: bytes, from_client: bool) -> tuple:
    x = 0 if from_client else 8

    sha1_a = sha1(msg_key + auth_key[x:x + 32]).digest()
    sha1_b = sha1(auth_key[32 + x:32 + x + 16] + msg_key + auth_key[48 + x:48 + x + 16]).digest()
    sha1_c = sha1(auth_key[64 + x:64 + x + 32] + msg_key).digest()
    sha1_d = sha1(msg_key + auth_key[96 + x:96 + x + 32]).digest()

    aes_key = sha1_a[0:8] + sha1_b[8:8 + 12] + sha1_c[4:4 + 12]
    aes_iv = sha1_a[8:8 + 12] + sha1_b[0:8] + sha1_c[16:16 + 4] + sha1_d[0:8]

    return aes_key, aes_iv


CtrTuple = tuple[bytes, bytes, bytearray]



from __future__ import annotations

from abc import ABC, abstractmethod
from os import urandom

from mtproto import transports
from mtproto.buffer import RxBuffer, TxBuffer
from mtproto.crypto.aes import ctr256_decrypt, ctr256_encrypt
from mtproto.enums import ConnectionRole
from mtproto.packets import BasePacket

HTTP_HEADER = {b"POST", b"GET ", b"HEAD", b"OPTI"}


class BaseTransport(ABC):
    __slots__ = ("our_role", "rx_buffer", "tx_buffer",)

    def __init__(self, role: ConnectionRole):
        self.our_role = role
        self.rx_buffer: RxBuffer | None = None
        self.tx_buffer: TxBuffer | None = None

    @abstractmethod
    def read(self) -> BasePacket | None: ...

    @abstractmethod
    def write(self, packet: BasePacket) -> None: ...

    @abstractmethod
    def has_packet(self) -> bool: ...

    def set_buffers(self, rx_buffer: RxBuffer, tx_buffer: TxBuffer) -> tuple[RxBuffer, TxBuffer]:
        self.rx_buffer = rx_buffer
        self.tx_buffer = tx_buffer

        return rx_buffer, tx_buffer

    @classmethod
    def from_buffer(cls, buf: RxBuffer, _four_ef: bool = False) -> BaseTransport | None:
        ef_count = 4 if _four_ef else 1
        if (header := buf.peekexactly(ef_count)) is None:
            return

        if header == b"\xef" * ef_count:
            buf.readexactly(ef_count)
            return transports.AbridgedTransport(ConnectionRole.SERVER)

        if (header := buf.peekexactly(4)) is None:
            return

        if header == b"\xee" * 4:
            buf.readexactly(4)
            return transports.IntermediateTransport(ConnectionRole.SERVER)
        elif header == b"\xdd" * 4:
            buf.readexactly(4)
            return transports.PaddedIntermediateTransport(ConnectionRole.SERVER)
        elif header in HTTP_HEADER:
            ...  # TODO: http transport
        elif buf.peekexactly(4, 4) == b"\x00" * 4:
            return transports.FullTransport(ConnectionRole.SERVER)
        elif buf.size() < 64:
            return

        nonce = buf.readexactly(64)
        temp = nonce[8:56][::-1]
        encrypt = (nonce[8:40], nonce[40:56], bytearray(1))
        decrypt = (temp[0:32], temp[32:48], bytearray(1))
        decrypted = ctr256_decrypt(nonce, *encrypt)
        header = decrypted[56:56 + 4]

        if (transport := cls.from_buffer(RxBuffer(header), True)) is None:
            raise ValueError(f"Unknown transport!")

        return transports.ObfuscatedTransport(transport, decrypt, encrypt)

    @classmethod
    def new(cls, buf: TxBuffer, transport_cls: type[BaseTransport], obf: bool) -> BaseTransport:
        if obf:
            if issubclass(transport_cls, transports.FullTransport):
                raise ValueError("Obfuscation of \"Full\" transport is nut supported")
            tmp_buf = TxBuffer()
            nonobf_transport = cls.new(tmp_buf, transport_cls, False)

            while True:
                nonce = bytearray(urandom(64))
                if (nonce[0] not in (0xef, 0xee, 0xdd)
                        and bytes(nonce[:4]) not in HTTP_HEADER
                        and nonce[4:8] != b"\x00" * 4):
                    nonce[56:60] = tmp_buf.data()[0:1] * 4
                    break

            temp = bytearray(nonce[55:7:-1])
            encrypt = (nonce[8:40], nonce[40:56], bytearray(1))
            decrypt = (temp[0:32], temp[32:48], bytearray(1))
            nonce[56:64] = ctr256_encrypt(nonce, *encrypt)[56:64]

            buf.write(bytes(nonce))
            return transports.ObfuscatedTransport(nonobf_transport, encrypt, decrypt)

        if issubclass(transport_cls, transports.AbridgedTransport):
            buf.write(b"\xef")
            return transports.AbridgedTransport(ConnectionRole.CLIENT)
        elif issubclass(transport_cls, transports.PaddedIntermediateTransport):
            buf.write(b"\xdd" * 4)
            return transports.PaddedIntermediateTransport(ConnectionRole.CLIENT)
        elif issubclass(transport_cls, transports.IntermediateTransport):
            buf.write(b"\xee" * 4)
            return transports.IntermediateTransport(ConnectionRole.CLIENT)
        elif issubclass(transport_cls, transports.FullTransport):
            return transports.FullTransport(ConnectionRole.CLIENT)

        raise ValueError(f"Unknown transport class: {transport_cls}")



from __future__ import annotations

from mtproto.crypto.aes import ctr256_decrypt, ctr256_encrypt, CtrTuple


class RxBuffer:
    __slots__ = ("_data",)

    def __init__(self, data: bytes = b""):
        self._data = data

    def size(self) -> int:
        return len(self._data)

    def readexactly(self, n: int) -> bytes | None:
        if self.size() < n:
            return

        data, self._data = self._data[:n], self._data[n:]

        return data

    def readall(self) -> bytes:
        data, self._data = self._data, b""
        return data

    def peekexactly(self, n: int, offset: int = 0) -> bytes | None:
        if self.size() < (n + offset):
            return

        return self._data[offset:offset+n]

    def data_received(self, data: bytes) -> None:
        self._data += data


class TxBuffer:
    __slots__ = ("_data",)

    def __init__(self, data: bytes = b""):
        self._data = data

    def data(self) -> bytes:
        return self._data

    def write(self, data: bytes | TxBuffer) -> None:
        if isinstance(data, TxBuffer):
            data = data.get_data()
        self._data += data

    def get_data(self) -> bytes:
        data, self._data = self._data, b""
        return data


class ObfuscatedRxBuffer(RxBuffer):
    __slots__ = ("_buffer", "_decrypt")

    def __init__(self, buffer: RxBuffer, decrypt: CtrTuple):
        super().__init__()

        self._buffer = buffer
        self._decrypt = decrypt

    def size(self) -> int:
        return self._buffer.size()

    def readexactly(self, n: int) -> bytes | None:
        return self._buffer.readexactly(n)

    def readall(self) -> bytes:
        return self._buffer.readall()

    def peekexactly(self, n: int, offset: int = 0) -> bytes | None:
        return self._buffer.peekexactly(n, offset)

    def data_received(self, data: bytes) -> None:
        if data:
            self._buffer.data_received(ctr256_decrypt(data, *self._decrypt))


class ObfuscatedTxBuffer(TxBuffer):
    __slots__ = ("_buffer", "_encrypt",)

    def __init__(self, buffer: TxBuffer, encrypt: CtrTuple):
        super().__init__()

        self._buffer = buffer
        self._encrypt = encrypt

    def data(self) -> bytes:
        return self._buffer.data()

    def write(self, data: bytes | TxBuffer) -> None:
        return self._buffer.write(data)

    def get_data(self) -> bytes:
        return ctr256_encrypt(self._buffer.get_data(), *self._encrypt)



from __future__ import annotations

from zlib import crc32

from .base_transport import BaseTransport
from ..buffer import TxBuffer
from ..packets import BasePacket, QuickAckPacket, ErrorPacket, MessagePacket


class FullTransport(BaseTransport):
    __slots__ = ("_seq_no_r", "_seq_no_w",)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._seq_no_r = self._seq_no_w = 0

    def read(self) -> BasePacket | None:
        if self.rx_buffer.size() < 4:
            return

        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little")
        if self.rx_buffer.size() < length:
            return

        length_bytes = self.rx_buffer.readexactly(4)
        seq_no_bytes = self.rx_buffer.readexactly(4)
        seq_no = int.from_bytes(seq_no_bytes, "little")
        data = self.rx_buffer.readexactly(length - 12)
        crc = int.from_bytes(self.rx_buffer.readexactly(4), "little")

        if crc != crc32(length_bytes + seq_no_bytes + data):
            return
        if seq_no != self._seq_no_r:
            return
        self._seq_no_r += 1

        if len(data) == 4:
            return ErrorPacket(int.from_bytes(data, "little", signed=True))

        return MessagePacket.parse(data, False)

    def write(self, packet: BasePacket) -> None:
        if isinstance(packet, QuickAckPacket):
            raise ValueError("\"Full\" transport does not support quick-acks.")

        data = packet.write()

        tmp = TxBuffer()
        tmp.write((len(data) + 12).to_bytes(4, byteorder="little"))
        tmp.write(self._seq_no_w.to_bytes(4, "little"))
        tmp.write(data)
        tmp.write(crc32(tmp.data()).to_bytes(4, byteorder="little"))

        self._seq_no_w += 1

        self.tx_buffer.write(tmp)

    def has_packet(self) -> bool:
        if self.rx_buffer.size() < 4:
            return False

        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little")
        return self.rx_buffer.size() >= length



from reprlib import recursive_repr


class AutoRepr:
    @recursive_repr()
    def __repr__(self) -> str:  # pragma: no cover
        cls = self.__class__
        add_fields = getattr(cls, "AUTOREPR_ADD_FIELDS", [])
        exclude_fields = set(getattr(cls, "AUTOREPR_EXCLUDE_FIELDS", []))

        fields_names = set(getattr(cls, "__slots__", self.__dict__.keys()))
        fields_names.update(*add_fields)
        fields_names.difference_update(*exclude_fields)

        fields = ", ".join([
            f"{name}={getattr(self, name, None)!r}"
            for name in fields_names
        ])

        return f"{cls.__name__}({fields})"



from __future__ import annotations

from .base_transport import BaseTransport
from ..buffer import TxBuffer, RxBuffer, ObfuscatedRxBuffer, ObfuscatedTxBuffer
from ..crypto.aes import CtrTuple
from ..packets import BasePacket


class ObfuscatedTransport(BaseTransport):
    __slots__ = ("_transport", "_encrypt", "_decrypt",)

    def __init__(self, transport: BaseTransport, encrypt: CtrTuple, decrypt: CtrTuple) -> None:
        super().__init__(transport.our_role)

        self._transport = transport
        self._encrypt = encrypt
        self._decrypt = decrypt

    def set_buffers(self, rx_buffer: RxBuffer, tx_buffer: TxBuffer) -> tuple[RxBuffer, TxBuffer]:
        back_rx, back_tx = RxBuffer(), TxBuffer()
        obf_rx = ObfuscatedRxBuffer(back_rx, self._decrypt)
        obf_tx = ObfuscatedTxBuffer(back_tx, self._encrypt)
        obf_rx.data_received(rx_buffer.readall())

        self._transport.set_buffers(back_rx, back_tx)

        return obf_rx, obf_tx

    def read(self) -> BasePacket | None:
        return self._transport.read()

    def write(self, packet: BasePacket) -> None:
        return self._transport.write(packet)

    def has_packet(self) -> bool:
        return self._transport.has_packet()



from .base_packet import BasePacket
from .error_packet import ErrorPacket
from .message_packet import MessagePacket, UnencryptedMessagePacket, EncryptedMessagePacket, DecryptedMessagePacket
from .quick_ack_packet import QuickAckPacket



from __future__ import annotations

from .buffer import RxBuffer, TxBuffer
from .enums import ConnectionRole
from .packets import BasePacket
from .transports import AbridgedTransport
from .transports.base_transport import BaseTransport


class Connection:
    __slots__ = ("_role", "_rx_buffer", "_tx_buffer", "_transport", "_transport_cls", "_transport_obf")

    def __init__(
            self,
            role: ConnectionRole = ConnectionRole.CLIENT,
            transport_cls: type[BaseTransport] = AbridgedTransport,
            transport_obf: bool = False,
    ):
        self._role = role
        self._rx_buffer = RxBuffer()
        self._tx_buffer = TxBuffer()
        self._transport: BaseTransport | None = None
        self._transport_cls = transport_cls
        self._transport_obf = transport_obf

    def receive(self, data: bytes = b"") -> BasePacket | None:
        self._rx_buffer.data_received(data)
        if self._transport is None and self._role == ConnectionRole.SERVER:
            self._transport = BaseTransport.from_buffer(self._rx_buffer)
            if self._transport is None:
                return
            self._rx_buffer, self._tx_buffer = self._transport.set_buffers(self._rx_buffer, self._tx_buffer)
        elif self._transport is None:
            raise ValueError("Transport should exist when receive() method is called and role is ConnectionRole.CLIENT")

        return self._transport.read()

    def send(self, packet: BasePacket) -> bytes:
        initial_data = b""
        if self._transport is None and self._role == ConnectionRole.CLIENT:
            init_buf = TxBuffer()
            self._transport = BaseTransport.new(init_buf, self._transport_cls, self._transport_obf)
            initial_data = init_buf.get_data()
            self._rx_buffer, self._tx_buffer = self._transport.set_buffers(self._rx_buffer, self._tx_buffer)
        elif self._transport is None:
            raise ValueError("Transport should exist when send() method is called and role is ConnectionRole.SERVER")

        self._transport.write(packet)
        return initial_data + self._tx_buffer.get_data()

    def has_packet(self) -> bool:
        return self._transport is not None and self._transport.has_packet()

    def opposite(self, require_transport: bool = True) -> Connection | None:
        if self._transport_cls is None:
            if require_transport:
                raise ValueError("transport_cls is required!")
            return

        return Connection(
            role=ConnectionRole.CLIENT if self._role is ConnectionRole.SERVER else ConnectionRole.SERVER,
            transport_cls=self._transport_cls,
            transport_obf=self._transport_obf,
        )



from mtproto.packets import BasePacket
from mtproto.utils import AutoRepr


class ErrorPacket(BasePacket, AutoRepr):
    __slots__ = ("error_code",)

    def __init__(self, error_code: int):
        self.error_code = abs(error_code)

    def write(self) -> bytes:
        return (-self.error_code).to_bytes(4, "little", signed=True)



from __future__ import annotations

from abc import ABC
from hashlib import sha256, sha1
from io import BytesIO
from os import urandom

from mtproto.crypto import kdf, ige256_encrypt, ige256_decrypt
from mtproto.crypto.aes import kdf_v1
from mtproto.enums import ConnectionRole
from mtproto.utils import AutoRepr
from .base_packet import BasePacket
from .quick_ack_packet import QuickAckPacket


class MessagePacket(BasePacket, ABC):
    @classmethod
    def parse(cls, payload: bytes, needs_quick_ack: bool = False) -> MessagePacket | None:
        buf = BytesIO(payload)
        auth_key_id = int.from_bytes(buf.read(8), "little")
        if auth_key_id == 0:
            message_id = int.from_bytes(buf.read(8), "little")
            message_length = int.from_bytes(buf.read(4), "little")
            return UnencryptedMessagePacket(message_id, buf.read(message_length))

        message_key = buf.read(16)
        encrypted_data = buf.read()
        return EncryptedMessagePacket(auth_key_id, message_key, encrypted_data, needs_quick_ack)


class UnencryptedMessagePacket(MessagePacket, AutoRepr):
    __slots__ = ("message_id", "message_data",)

    def __init__(self, message_id: int, message_data: bytes):
        self.message_id = message_id
        self.message_data = message_data

    def write(self) -> bytes:
        return (
                (0).to_bytes(8, "little") +
                self.message_id.to_bytes(8, "little") +
                len(self.message_data).to_bytes(4, "little") +
                self.message_data
        )


class EncryptedMessagePacket(MessagePacket, AutoRepr):
    __slots__ = ("auth_key_id", "message_key", "encrypted_data", "needs_quick_ack",)

    def __init__(self, auth_key_id: int, message_key: bytes, encrypted_data: bytes, needs_quick_ack: bool = False):
        self.auth_key_id = auth_key_id
        self.message_key = message_key
        self.encrypted_data = encrypted_data
        self.needs_quick_ack = needs_quick_ack

    def write(self) -> bytes:
        return (
                self.auth_key_id.to_bytes(8, "little") +
                self.message_key +
                self.encrypted_data
        )

    def decrypt(self, auth_key: bytes, sender_role: ConnectionRole, v1: bool = False) -> DecryptedMessagePacket:
        if (got_key_id := int.from_bytes(sha1(auth_key).digest()[-8:], "little")) != self.auth_key_id:
            raise ValueError(f"Invalid auth_key: expected key with id {self.auth_key_id}, got {got_key_id}")

        kdf_func = kdf_v1 if v1 else kdf
        aes_key, aes_iv = kdf_func(auth_key, self.message_key, sender_role == ConnectionRole.CLIENT)

        decrypted = ige256_decrypt(self.encrypted_data, aes_key, aes_iv)
        return DecryptedMessagePacket.parse(decrypted)


class DecryptedMessagePacket(MessagePacket, AutoRepr):
    __slots__ = ("salt", "session_id", "message_id", "seq_no", "data", "padding",)

    def __init__(
            self, salt: bytes, session_id: int, message_id: int, seq_no: int, data: bytes, padding: bytes | None = None
    ):
        self.salt = salt
        self.session_id = session_id
        self.message_id = message_id
        self.seq_no = seq_no
        self.data = data
        self.padding = padding

    def write(self) -> bytes:
        raise NotImplementedError(
            f"{self.__class__.__name__}.write is not implemented. "
            f"You should call {self.__class__.__name__}.encrypt and call .write on returned encrypted message."
        )

    @classmethod
    def parse(cls, data: bytes, *args, **kwargs) -> DecryptedMessagePacket:
        buf = BytesIO(data)
        salt = buf.read(8)
        session_id = int.from_bytes(buf.read(8), "little")
        message_id = int.from_bytes(buf.read(8), "little")
        seq_no = int.from_bytes(buf.read(4), "little")
        length = int.from_bytes(buf.read(4), "little")

        return cls(
            salt,
            session_id,
            message_id,
            seq_no,
            buf.read(length),
            buf.read(),
        )

    def encrypt(self, auth_key: bytes, sender_role: ConnectionRole) -> EncryptedMessagePacket:
        data = (
                self.salt +
                self.session_id.to_bytes(8, "little") +
                self.message_id.to_bytes(8, "little") +
                self.seq_no.to_bytes(4, "little") +
                len(self.data).to_bytes(4, "little") +
                self.data
        )

        padding = urandom(-(len(data) + 12) % 16 + 12)

        # 96 = 88 + 8 (8 = incoming message (server message); 0 = outgoing (client message))
        key_offset = 88 + (0 if sender_role == ConnectionRole.CLIENT else 8)
        msg_key_large = sha256(auth_key[key_offset:key_offset + 32] + data + padding).digest()
        msg_key = msg_key_large[8:24]
        aes_key, aes_iv = kdf(auth_key, msg_key, sender_role == ConnectionRole.CLIENT)

        return EncryptedMessagePacket(
            int.from_bytes(sha1(auth_key).digest()[-8:], "little"),
            msg_key,
            ige256_encrypt(data + padding, aes_key, aes_iv),
        )

    def quick_ack_response(self, auth_key: bytes, sender_role: ConnectionRole) -> QuickAckPacket:
        key_offset = 88 + (0 if sender_role == ConnectionRole.CLIENT else 8)
        msg_key_large = sha256(auth_key[key_offset:key_offset + 32] + self.data + self.padding).digest()

        return QuickAckPacket(
            (int.from_bytes(msg_key_large[:4], "little") | 0x80000000).to_bytes(4, "little")
        )



from .abridged import AbridgedTransport
from .full import FullTransport
from .intermediate import IntermediateTransport
from .obfuscated import ObfuscatedTransport
from .padded_intermediate import PaddedIntermediateTransport



from mtproto.packets import BasePacket
from mtproto.utils import AutoRepr


class QuickAckPacket(BasePacket, AutoRepr):
    __slots__ = ("token",)

    def __init__(self, token: bytes):
        self.token = token

    def write(self) -> bytes:
        return self.token



from __future__ import annotations

from .base_transport import BaseTransport
from ..enums import ConnectionRole
from ..packets import BasePacket, QuickAckPacket, ErrorPacket, MessagePacket


class IntermediateTransport(BaseTransport):
    def read(self) -> BasePacket | None:
        if self.rx_buffer.size() < 4:
            return

        is_quick_ack = (self.rx_buffer.peekexactly(1)[0] & 0x80) == 0x80
        if is_quick_ack and self.our_role == ConnectionRole.CLIENT:
            return QuickAckPacket(self.rx_buffer.readexactly(4))

        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little") & 0x7FFFFFFF
        if self.rx_buffer.size() < length:
            return

        self.rx_buffer.readexactly(4)
        data = self.rx_buffer.readexactly(length)
        if len(data) == 4:
            return ErrorPacket(int.from_bytes(data, "little", signed=True))

        return MessagePacket.parse(data, is_quick_ack)

    def write(self, packet: BasePacket) -> None:
        data = packet.write()
        if isinstance(packet, QuickAckPacket):
            self.tx_buffer.write(data)
            return

        self.tx_buffer.write(len(data).to_bytes(4, byteorder="little"))
        self.tx_buffer.write(data)

    def has_packet(self) -> bool:
        if self.rx_buffer.size() < 4:
            return False
        if self.rx_buffer.peekexactly(1)[0] & 0x80 == 0x80:  # TODO: ?
            return True

        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little") & 0x7FFFFFFF
        return self.rx_buffer.size() >= (length + 4)



from __future__ import annotations

import os
from random import randint

from . import IntermediateTransport
from ..packets import BasePacket, QuickAckPacket, MessagePacket, ErrorPacket


class PaddedIntermediateTransport(IntermediateTransport):
    def read(self) -> BasePacket | None:
        if self.rx_buffer.size() < 4:
            return

        is_quick_ack = (self.rx_buffer.peekexactly(1)[0] & 0x80) == 0x80
        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little") & 0x7FFFFFFF
        if self.rx_buffer.size() < length:
            return

        self.rx_buffer.readexactly(4)
        data = self.rx_buffer.readexactly(length)
        if length > 16:
            return MessagePacket.parse(
                data[:(length - length % 4)],
                is_quick_ack,
            )

        if data[:4] == b"\xff\xff\xff\xff":  # TODO: is check for self.role == ConnectionRole.CLIENT needed?
            return QuickAckPacket(data[4:8])

        return ErrorPacket(int.from_bytes(data[:4], "little", signed=True))

    def write(self, packet: BasePacket) -> None:
        data = packet.write()
        if isinstance(packet, QuickAckPacket):
            data = b"\xff\xff\xff\xff" + data

        data += os.urandom(randint(0, 3))
        self.tx_buffer.write(len(data).to_bytes(4, byteorder="little"))
        self.tx_buffer.write(data)

    def has_packet(self) -> bool:
        if self.rx_buffer.size() < 4:
            return False

        length = int.from_bytes(self.rx_buffer.peekexactly(4), "little") & 0x7FFFFFFF
        return self.rx_buffer.size() >= (length + 4)
